#!/usr/bin/env python3

import pickle
from typing import List
from functools import lru_cache

import numpy as np

from .optimizers import Adam, Optimizer


@lru_cache(maxsize=1)
def utility_function(population_size: int) -> List[float]:
    utility = np.clip(
        np.log(population_size/2 + 1) -
        np.log(np.arange(1, population_size+1)),
        0, None)
    utility /= np.sum(
        np.clip(np.log(population_size/2 + 1) -
                np.log(np.arange(1, population_size+1)),
                0, None))
    utility -= 1 / population_size
    return utility

def fitness_transformation(fitness_scores: List[float]) -> List[float]:
    ranks = np.argsort(fitness_scores)[::-1]
    population_size = len(fitness_scores)
    utility = utility_function(population_size)
    return utility[ranks]


class PEPG():
    def __init__(self, population_size: int, theta_size: int,
                 mu_init: float, sigma_init: float, mu_lr: float,
                 sigma_lr: float, l2_coeff: float = 0.005,
                 optimizer = Adam, optimizer_kwargs = {}):
        self.population_size: int = population_size
        self.theta_size: int = theta_size
        self.mu_init: float = mu_init
        self.sigma_init: float = sigma_init
        self.mu_lr: float = mu_lr
        self.sigma_lr: float = sigma_lr
        self.l2_coeff: float = l2_coeff

        self.step = 0

        self.mu = mu_init * np.ones(theta_size)
        self.sigma = sigma_init * np.ones(theta_size)

        self.best_fitness = -float('inf')
        self.best_mu = None
        self.best_sigma = None
        self.best_theta = None
            
        assert issubclass(optimizer, Optimizer)
        self.optimizer = optimizer(mu_lr, **optimizer_kwargs)

    def get_parameters(self) -> List[List[float]]:
        self.epsilon = np.random.normal(
            0.0, self.sigma,
            (int(self.population_size/2), self.mu.size))
        self.theta = np.concatenate([self.mu + self.epsilon, self.mu - self.epsilon])
        return self.theta

    def update(self, rewards):
        idx = np.argmax(rewards)
        if rewards[idx] > self.best_fitness:
            self.best_mu = self.mu
            self.best_sigma = self.sigma
            self.best_theta = self.theta[idx]
            self.best_fitness = rewards[idx]

        rewards = fitness_transformation(rewards)
        baseline = np.mean(rewards)
        batch = int(self.population_size/2)
        s = ((self.epsilon ** 2) - (self.sigma ** 2)) / self.sigma
        rt = rewards[:batch] - rewards[batch:]
        rs = ((rewards[:batch] + rewards[batch:]) / 2 - baseline)
        mu_g = np.dot(self.epsilon.T, rt) / self.population_size
        self.mu += self.optimizer(-mu_g + self.l2_coeff * self.mu)
        self.sigma += self.sigma_lr * np.dot(s.T, rs) / self.population_size
        self.step += 1

    def save_checkpoint(self):
        import time
        import pathlib
        import os
        import inspect
        frame = inspect.stack()[1]
        module = inspect.getmodule(frame[0])
        folder = pathlib.Path(
            f"{os.path.join(os.path.dirname(module.__file__), 'checkpoints')}")
        folder.mkdir(parents=True, exist_ok=True)
        filename = f'{int(time.time())}.checkpoint'
        save_path = os.path.abspath(os.path.join(folder, filename))
        print(f'Saving checkpoint: {save_path}')
        with open(os.path.join(folder, filename), 'wb') as output:
            pickle.dump(self.__dict__, output, -1)

    @classmethod
    def load_checkpoint(cls, filename):
        with open(filename, 'rb') as checkpoint:
            cls_dict = pickle.load(checkpoint)
        es = cls.__new__(cls)
        es.__dict__.update(cls_dict)
        return es

    def save_best(self, filename):
        with open(filename, 'wb') as output:
            pickle.dump((self.best_theta, self.best_mu, self.best_sigma), output, -1)

    @classmethod
    def load_best(cls, filename):
        with open(filename, 'rb') as best:
            theta, mu, sigma = pickle.load(best)
        return theta, mu, sigma



#!/usr/bin/env python

import os
from setuptools import setup

directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='pepg-es',
      version='0.0.3',
      description='Python Implementation of Parameter-exploring Policy Gradients Evolution Strategy',
      author='Göktuğ Karakaşlı',
      author_email='karakasligk@gmail.com',
      license='MIT',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://github.com/goktug97/PEPG-ES',
      download_url=(
          'https://github.com/goktug97/PEPG-ES/archive/v0.0.2.tar.gz'),
      packages = ['pepg'],
      classifiers=[
          "Programming Language :: Python :: 3",
          "License :: OSI Approved :: MIT License",
          "Operating System :: POSIX :: Linux",
      ],
      install_requires=[
          'numpy'
      ],
      python_requires='>=3.6',
      include_package_data=True)



import pickle
from typing import Callable, List

import numpy as np


def sigmoid(x):
    return 1 / (1 + np.exp(-x))

def tanh(x):
    return np.tanh(x)


class NeuralNetwork(object):
    def __init__(self, input_size:int, output_size: int, hidden_sizes: List[int] = [],
                 hidden_activation: Callable[[float], float] = lambda x: x,
                 output_activation: Callable[[float], float] = lambda x: x,
                 bias: bool = True):
        self.input_size = input_size
        self.output_size = output_size
        self.hidden_sizes = hidden_sizes
        self.hidden_activation = hidden_activation
        self.output_activation = output_activation
        self.bias = bias

        sizes = np.concatenate([[input_size], hidden_sizes, [output_size]])
        number_of_parameters = 0
        for i in range(len(sizes)-1):
            number_of_parameters += sizes[i] * sizes[i+1]
        if bias:
            number_of_parameters += np.sum(sizes[1:])
        self.number_of_parameters: int = int(number_of_parameters)
        self.sizes = sizes.astype('int')

        self._weights = None

    @property
    def weights(self):
        return self._weights

    @weights.setter
    def weights(self, value):
        self._weights = np.array(value)

    def __call__(self, value):
        assert self._weights is not None

        weights = self.weights
        output = value
        if self.bias:
            output = np.append(output, 1)
        index = 0
        for i in range(len(self.sizes)-2):
            w = weights[index:index+self.sizes[i]*self.sizes[i+1]+self.sizes[i+1]]
            w = w.reshape(self.sizes[i]+1, self.sizes[i+1])
            index += self.sizes[i]*self.sizes[i+1]+self.sizes[i+1]
            output = self.hidden_activation(output @ w)
            if self.bias:
                output = np.append(output, 1)

        w = weights[index:]
        w = w.reshape(self.sizes[len(self.sizes)-2]+(1 if self.bias else 0),
                      self.sizes[len(self.sizes)-1])
        output = self.output_activation(output @ w)
        return output

    def save_network(self, filename):
        with open(filename, 'wb') as output:
            pickle.dump(self.__dict__, output, -1)

    @classmethod
    def load_network(cls, filename):
        with open(filename, 'rb') as f:
            cls_dict = pickle.load(f)
        network = cls.__new__(cls)
        network.__dict__.update(cls_dict)
        return network



from abc import ABC, abstractmethod

import numpy as np

class Optimizer(ABC):
    @abstractmethod
    def __init__(self, alpha):
        pass

    @abstractmethod
    def __call__(self, gradients):
        pass

class Adam(Optimizer):
    def __init__(self, alpha: float, beta_1: float = 0.9, beta_2:
                 float = 0.999, epsilon: float = 1e-08):

        self.alpha = alpha
        self.beta_1 = beta_1
        self.beta_2 = beta_2
        self.epsilon = epsilon

        self.m = None
        self.v = None

        self.t = 0

    def __call__(self, gradients):
        self.t += 1
        if self.m is None:
            self.m = np.zeros_like(gradients)
            self.v = np.zeros_like(gradients)
        alpha = self.alpha * (np.sqrt(1 - np.power(self.beta_2, self.t)) /
                              (1 - np.power(self.beta_1, self.t)))
        self.m = self.beta_1 * self.m + (1 - self.beta_1) * gradients
        self.v = self.beta_2 * self.v + (1 - self.beta_2) * np.square(gradients)
        return -alpha * self.m / (np.sqrt(self.v) + self.epsilon)

